import json
import os
import pickle

import cv2
import h5py
import imageio
import numpy as np
import scipy.io
import yaml


def load(path):
  ext = os.path.splitext(path)[1][1:]
  if ext == 'json':
    return load_json(path)
  elif ext == 'mat':
    return load_mat(path)
  elif ext == 'h5':
    return load_hdf5(path)
  elif ext == 'txt':
    return load_txt(path)
  elif ext == 'pkl':
    return load_pkl(path)
  elif ext in ['png', 'jpg', 'jpeg']:
    return np.array(load_img(path))
  elif ext == 'obj':
    return load_obj(path)
  elif ext == 'npy' or ext == 'npz':
    return np.load(path)
  elif ext == 'yaml':
    return load_yaml(path)
  else:
    raise NotImplementedError('Unsupported file extension: ' + path)


def save(path, data, **kwargs):
  ext = os.path.splitext(path)[1][1:]
  if ext == 'json':
    return save_json(path, data, **kwargs)
  elif ext == 'h5':
    return save_hdf5(path, data, **kwargs)
  elif ext == 'txt':
    return save_txt(path, data, **kwargs)
  elif ext == 'pkl':
    return save_pkl(path, data, **kwargs)
  elif ext in ['png', 'jpg', 'jpeg']:
    return save_img(path, data, **kwargs)
  elif ext == 'obj':
    return save_obj(path, *data, **kwargs)
  elif ext == 'npy':
    return np.save(path, data, **kwargs)
  elif ext == 'npz':
    return np.savez(path, data, **kwargs)
  elif ext == 'yaml':
    return save_yaml(path, data, **kwargs)
  elif ext == 'gif':
    imageio.mimsave(path, data, **kwargs)
  elif ext in ['mp4', 'avi']:
    return save_video(path, data, **kwargs)
  else:
    raise NotImplementedError('Unsupported file extension: ' + path)


def save_video(path, data, fps):
  writer = VideoWriter(path, data[0].shape[1], data[0].shape[0], fps)
  for frame in data:
    writer.write_frame(frame)
  writer.close()


def load_yaml(path):
  with open(path) as f:
    data = yaml.safe_load(f)
  return data


def save_yaml(path, data):
  with open(path, 'w') as f:
    yaml.dump(data, f)


def load_json(path):
  """
  Load json data from file.

  Parameters
  ----------
  path : str
    Path to the json file.

  Returns
  -------
  dict or list
    Json data.
  """
  with open(path) as f:
    data = json.load(f)
  return data


def save_json(path, data):
  """
  Save dict data into json file.

  Parameters
  ----------
  path : str
    Path to save.
  data : dict or list
    Data to save.
  """
  with open(path, 'w') as f:
    f.write(json.dumps(data, indent=2))


def load_mat(path):
  """
  Load matlab .mat data.

  Parameters
  ----------
  path : str
    Path to data.

  Returns
  -------
  obj
    Data.
  """
  return scipy.io.loadmat(path)


def save_hdf5(path, data):
  """
  Save data into hdf5 format. If any data is string in a numpy array, make sure
  set the dtype to 'S'.

  Parameters
  ----------
  path : str
    Path to save the data.
  data : dict
    Keys and data.
  """
  f = h5py.File(path, 'w')
  for k, v in data.items():
    f.create_dataset(k, v.shape, data=v)
  f.close()


def load_hdf5(path):
  """
  Load data from hdf5 file.

  Parameters
  ----------
  path : str
    Path to the file.

  Returns
  -------
  dict
    Data.
  """
  f = h5py.File(path, 'r')
  data = {}
  for k in f.keys():
    data[k] = np.array(f[k])
  return data


def save_txt(path, s):
  with open(path, 'w') as f:
    f.write(s)


def load_txt(path):
  """
  Read all lines from a text file.

  Parameters
  ----------
  path : str
    Path to the text file.

  Returns
  -------
  list
    A list of lines.
  """
  with open(path, 'r') as f:
    lines = f.read().splitlines()
  return lines


def pkl_load(path):
  """
  Load pickle data.

  Parameters
  ----------
  path : str
    Path to the file.

  Returns
  -------
  object
    Loaded data.
  """

  with open(path, 'rb') as f:
    data = pickle.load(f)
  return data


load_pkl = pkl_load


def pkl_save(path, data):
  """
  Save pickle data.

  Parameters
  ----------
  path : str
    Path to save file.
  data : object
    Data to save.
  """
  with open(path, 'wb') as f:
    pickle.dump(data, f)


save_pkl = pkl_save


def obj_save(path, vertices, faces=None):
  """
  Save .obj mesh file.

  Parameters
  ----------
  path : str
    Path to save the file.
  vertices : np.ndarray
    Mesh vertices.
  faces : np.ndarray, optional
    Mesh faces, by default None
  """
  with open(path, 'w') as fp:
    for v in vertices:
      v = np.ravel(v)
      fp.write('v %f %f %f\n' % (v[0], v[1], v[2]))
    if faces is not None:
      for f in faces + 1:
        f = np.ravel(f)
        fp.write('f %d %d %d\n' % (f[0], f[1], f[2]))


save_obj = obj_save


def imread(path):
  """
  Read image.

  Parameters
  ----------
  path : str
    Path to read the image.

  Returns
  -------
  np.ndarray
    Read image.
  """
  return imageio.imread(path)


load_img = imread


def imsave(path, img):
  """
  Save image.

  Parameters
  ----------
  path : str
    Path to load image.
  img : np.ndarray
    Image to save.
  """
  imageio.imsave(path, img)


save_img = imsave


class VideoReader:
  """
  Read frames from video.

  Call `next_frame` to read next frame.
  Call `all_frames` to read all remaining frames.
  Call `close` to release resource.

  """
  def __init__(self, path):
    """
    Parameters
    ----------
    path : str
      Path to the video file.
    """
    self.video = cv2.VideoCapture(path)
    self.fps = self.video.get(cv2.CAP_PROP_FPS)
    self.width = int(self.video.get(cv2.CAP_PROP_FRAME_WIDTH))
    self.height = int(self.video.get(cv2.CAP_PROP_FRAME_HEIGHT))
    self.n_frames = int(self.video.get(cv2.CAP_PROP_FRAME_COUNT)) + 1
    self.frame_idx = 0

  def next_frame(self):
    """
    Read next frame. Return None if no frame left.

    Returns
    -------
    np.ndarray
      Image in RGB format.
    """
    if not self.video.isOpened():
      return None
    ret, frame = self.video.read()
    self.frame_idx += 1
    if not ret:
      return None
    frame = np.flip(frame, axis=-1).copy()
    return frame

  def all_frames(self):
    """
    Read all remaining frames.

    Returns
    -------
    list
      A list of frames.
    """
    frames = []
    while self.video.isOpened():
      ret, frame = self.video.read()
      self.frame_idx += 1
      if not ret:
        break
      frame = np.flip(frame, axis=-1).copy()
      frames.append(frame)
    return frames

  def close(self):
    """
    Release video resource.
    """
    self.video.release()


class VideoWriter:
  """
  Write frames to a video.

  Call `write_frame` to write a single frame.
  Call `close` to release resource.

  """
  def __init__(self, path, width, height, fps, codec='avc1'):
    """
    Parameters
    ----------
    path : str
      Path to the video.
    width : int
      Frame width.
    height : int
      Frame height.
    fps : int
      Video frame rate.
    codec : str, optional
      Video codec, by default H264.
    """
    self.fps = fps
    self.width = width
    self.height = height
    self.video = cv2.VideoWriter(
      path, cv2.VideoWriter_fourcc(*codec), fps, (width, height)
    )
    self.frame_idx = 0

  def write_frame(self, frame):
    """
    Write one frame.

    Parameters
    ----------
    frame : np.ndarray
      Frame to write.
    """
    self.video.write(np.flip(frame, axis=-1).copy())
    self.frame_idx += 1

  def close(self):
    """
    Release resource.
    """
    self.video.release()


def ply_save_color_face(path, verts, faces, colors):
  """
  Save mesh into a .ply file with colored faces.

  Parameters
  ----------
  path : str
    Path to save the file.
  verts : np.ndarray, shape [v, 3]
    Mesh vertices.
  faces : np.ndarray, shape [f, 3]
    Mesh faces.
  colors : np.ndarray, shape [f, 3], dtype uint8
    RGB color for each face.
  """
  num_verts = verts.shape[0]
  num_faces = faces.shape[0]
  with open(path, 'wb') as f:
    f.write(b'ply\n')
    f.write(b'format ascii 1.0\n')
    f.write(b'element vertex %d\n' % num_verts)
    f.write(b'property float32 x\n')
    f.write(b'property float32 y\n')
    f.write(b'property float32 z\n')
    f.write(b'element face %d\n' % num_faces)
    f.write(b'property list uint8 int32 vertex_index\n')
    f.write(b'property uchar red\n')
    f.write(b'property uchar green\n')
    f.write(b'property uchar blue\n')
    f.write(b'end_header\n')
    for i in range(num_verts):
      v = verts[i]
      f.write(b'%f %f %f\n' % (v[0], v[1], v[2]))
    for i in range(num_faces):
      face = faces[i]
      c = colors[i]
      f.write(
        b'3 %d %d %d %d %d %d\n' % (face[0], face[1], face[2], c[0], c[1], c[2])
      )


save_ply_color_face = ply_save_color_face


def ply_save_color_pcloud(path, pcloud, color):
  """
  Save point cloud into a .ply file with per-point color.

  Parameters
  ----------
  path : str
    Path to save the file.
  pcloud : np.ndarray, shape [n, 3]
    Point cloud to save.
  color : np.ndarray, shape [n, 3], dtype uint8
    Color for each point.
  """
  num_points = pcloud.shape[0]
  with open(path, 'wb') as f:
    f.write(b'ply\n')
    f.write(b'format ascii 1.0\n')
    f.write(b'element vertex %d\n' % num_points)
    f.write(b'property float32 x\n')
    f.write(b'property float32 y\n')
    f.write(b'property float32 z\n')
    f.write(b'property uchar red\n')
    f.write(b'property uchar green\n')
    f.write(b'property uchar blue\n')
    f.write(b'end_header\n')
    for i in range(num_points):
      v = pcloud[i]
      c = color[i]
      f.write(b'%f %f %f %d %d %d\n' % (v[0], v[1], v[2], c[0], c[1], c[2]))


save_ply_color_pcloud = ply_save_color_pcloud


def load_obj(file_path):
  """
  Read the most naive .obj file. Only supports `v` and `f`.

  Parameters
  ----------
  file_path : str
    Path to the file.

  Returns
  -------
  [N, 3]
    Vertices.
  [N, 3]
    Faces (Vertex index starts from 0).
  """
  lines = load_txt(file_path)
  verts = []
  faces = []
  for l in lines:
    units = l.split()
    if len(units) != 4:
      continue
    if units[0] == 'v':
      verts.append(np.array([float(v) for v in units[1:]]))
    elif units[0] == 'f':
      faces.append(np.array([int(f.split('/')[0]) - 1 for f in units[1:]]))
  verts = np.array(verts)
  faces = np.array(faces)
  return verts, faces



import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()
import numpy as np


def tensor_shape(t):
  """
  Get the shape of tensor.

  Parameters
  ----------
  t : tensor
    Input tensor.

  Returns
  -------
  list
    Shape of tensor.
  """
  return t.get_shape().as_list()


def conv(inputs, oc, ks, st, rt=1, pd='SAME'):
  """
  Convolution.

  Parameters
  ----------
  inputs : tensor
    Input tensor.
  oc : int
    Number of output channels.
  ks : int
    Kernel size.
  st : int
    Stride.
  rt : int, optional
    Dilation rate, by default 1.
  pd : str
    Padding, by default 'SAME'.

  Returns
  -------
  tensor
    Output tensor.
  """
  layer = tf.layers.conv2d(
    inputs, oc, ks, strides=st, padding=pd, use_bias=False, dilation_rate=rt
  )
  return layer


def conv_bn(inputs, oc, ks, st, scope, training, rt=1, pd='SAME'):
  """
  Convolution - batch normalization.

  Parameters
  ----------
  inputs : tensor
    Input tensor.
  oc : int
    Number of output channels.
  ks : int
    Kernel size.
  st : int
    Stride.
  scope : str
    Variable scope.
  training : bool
    Is training or not.
  rt : int, optional
    Dilation rate, by default 1.
  pd : str
    Padding, by default 'SAME'.

  Returns
  -------
  tensor
    Output tensor.
  """
  with tf.variable_scope(scope):
    layer = conv(inputs, oc, ks, st, rt, pd)
    layer = tf.layers.batch_normalization(layer, training=training)
  return layer


def conv_bn_relu(inputs, oc, ks, st, scope, training, rt=1, pd='SAME'):
  """
  Convolution - batch normalization - relu.

  Parameters
  ----------
  inputs : tensor
    Input tensor.
  oc : int
    Number of output channels.
  ks : int
    Kernel size.
  st : int
    Stride.
  scope : str
    Variable scope.
  training : bool
    Training or not.
  rt : int, optional
    Dilation rate, by default 1.
  pd : str
    Padding, by default 'SAME'.


  Returns
  -------
  tensor
    Output tensor.
  """
  layer = conv_bn(inputs, oc, ks, st, scope, training, rt, pd)
  layer = tf.nn.relu(layer)
  return layer


def deconv(inputs, oc, ks, st):
  """
  Deconvolution.

  Parameters
  ----------
  inputs : tensor
    Input tensor.
  oc : int
    Number of output channels.
  ks : int
    Kernel size.
  st : int
    Stride.

  Returns
  -------
  tensor
    Output tensor.
  """
  layer = tf.layers.conv2d_transpose(
    inputs, oc, ks, st, padding='SAME', use_bias=False
  )
  return layer


def deconv_bn(inputs, oc, ks, st, scope, training):
  """
  Deconvolution - batch normalization.

  Parameters
  ----------
  inputs : tensor
    Input tensor.
  oc : int
    Number of output channels.
  ks : int
    Kernel size.
  st : int
    Stride.
  scope : str
    Variable scope.
  training : bool
    Is training or not.

  Returns
  -------
  tensor
    Output tensor.
  """
  with tf.variable_scope(scope):
    layer = deconv(inputs, oc, ks, st)
    layer = tf.layers.batch_normalization(layer, training=training)
  return layer


def deconv_bn_relu(inputs, oc, ks, st, scope, training):
  """
  Deconvolution - batch normalization - relu.

  Parameters
  ----------
  inputs : tensor
    Input tensor.
  oc : int
    Number of output channels.
  ks : int
    Kernel size.
  st : int
    Stride.
  scope : str
    Variable scope.
  training : bool
    Training or not.

  Returns
  -------
  tensor
    Output tensor.
  """
  layer = deconv_bn(inputs, oc, ks, st, scope, training)
  layer = tf.nn.relu(layer)
  return layer


def dense(layer, n_units):
  layer = tf.layers.dense(
    layer, n_units, activation=None,
    kernel_initializer=tf.initializers.truncated_normal(stddev=0.01)
  )
  return layer


def dense_bn(layer, n_units, scope, training):
  with tf.variable_scope(scope):
    layer = dense(layer, n_units)
    layer = tf.layers.batch_normalization(layer, training=training)
  return layer


def bottleneck(inputs, oc, st, scope, training, rate=1):
  """
  Bottleneck block for ResNet.

  Parameters
  ----------
  inputs : tensor
    Input tensor.
  oc : int
    Number of output channels.
  st : int
    Stride.
  scope : str
    Variable scope.
  training : bool
    Training or not.
  rate : int, optional
    Dilation rate, by default 1

  Returns
  -------
  tensor
    Output tensor.
  """
  with tf.variable_scope(scope):
    ic = inputs.get_shape().as_list()[-1]
    if ic == oc:
      if st == 1:
        shortcut = inputs
      else:
        shortcut = \
          tf.nn.max_pool2d(inputs, [1, st, st, 1], [1, st, st, 1], 'SAME')
    else:
      shortcut = conv_bn(inputs, oc, 1, st, 'shortcut', training)

    residual = conv_bn_relu(inputs, oc//4, 1, 1, 'conv1', training)
    residual = conv_bn_relu(residual, oc//4, 3, st, 'conv2', training, rate)
    residual = conv_bn(residual, oc, 1, 1, 'conv3', training)
    output = tf.nn.relu(shortcut + residual)

  return output


def resnet50(inputs, scope, training, squeeze, size=[3, 4, 6]):
  """
  ResNet 50.

  Parameters
  ----------
  inputs : tensor
    Input tensor.
  scope : str
    Variable scope.
  training : bool
    Training or not.
  squeeze: bool
    Squeeze 1024 channels to 256 channels in the end, or not.
  size: list of 3 numbers
    Number of layers for each block.

  Returns
  -------
  tensor
    Output tensor. If `internal_layer == True`, will also return that layer as
    the second return value.
  """
  with tf.variable_scope(scope):
    layer = conv_bn_relu(inputs, 64, 7, 2, 'conv1', training)

    with tf.variable_scope('block1'):
      for unit in range(size[0] - 1):
        layer = bottleneck(layer, 256, 1, 'unit%d' % (unit + 1), training)
      layer = bottleneck(layer, 256, 2, 'unit3', training)

    with tf.variable_scope('block2'):
      for unit in range(size[1]):
        layer = bottleneck(layer, 512, 1, 'unit%d' % (unit + 1), training, 2)

    with tf.variable_scope('block3'):
      for unit in range(size[2]):
        layer = bottleneck(layer, 1024, 1, 'unit%d' % (unit + 1), training, 4)

    if squeeze:
      layer = conv_bn_relu(layer, 256, 3, 1, 'squeeze', training)

  return layer


def hmap_to_uv(hmap):
  """
  Compute uv coordinates of heat map optima.

  Parameters
  ----------
  hmap : tensor, shape [n, h, w, c]
    Input heat maps.

  Returns
  -------
  tensor, shape [n, c, 2]
    UV coordinates, order (row, column).
  """
  hmap_flat = tf.reshape(hmap, (tf.shape(hmap)[0], -1, tf.shape(hmap)[3]))
  argmax = tf.argmax(hmap_flat, axis=1, output_type=tf.int32)
  argmax_r = argmax // tf.shape(hmap)[2]
  argmax_c = argmax % tf.shape(hmap)[2]
  uv = tf.stack([argmax_r, argmax_c], axis=2)
  return uv



import open3d as o3d
import numpy as np
from ..io import VideoWriter
from ..misc import progress_bar
from tqdm import tqdm


def render_sequence_3d(verts, faces, width, height, video_path, fps=30,
                       visible=False, need_norm=True, verbose=True, colors=None):
  """
  Render mesh animation using open3d.

  Parameters
  ----------
  verts : np.ndarray, shape [n, v, 3]
    Mesh vertices for each frame.
  faces : np.ndarray, shape [f, 3]
    Mesh faces.
  width : int
    Width of video.
  height : int
    Height of video.
  video_path : str
    Path to save the rendered video.
  fps : int, optional
    Video framerate, by default 30
  visible : bool, optional
    Wether to display rendering window, by default False
  need_norm : bool, optional
    Normalizing the vertices and locate camera automatically or not, by default
    True
  """
  if need_norm:
    if type(verts) == list:
      verts = np.stack(verts, 0)

    mean = np.mean(verts, axis=(0, 1), keepdims=True)
    scale = np.max(np.std(verts, axis=(0, 1), keepdims=True)) * 6
    verts = (verts - mean) / scale

  cam_offset = 1.2

  writer = VideoWriter(video_path, width, height, fps)

  mesh = o3d.geometry.TriangleMesh()
  mesh.triangles = o3d.utility.Vector3iVector(faces)
  mesh.vertices = o3d.utility.Vector3dVector(verts[0])

  vis = o3d.visualization.Visualizer()
  vis.create_window(width=width, height=height, visible=visible)
  vis.add_geometry(mesh)
  view_control = vis.get_view_control()
  cam_params = view_control.convert_to_pinhole_camera_parameters()
  cam_params.extrinsic = np.array([
    [1, 0, 0, 0],
    [0, 1, 0, 0],
    [0, 0, 1, cam_offset],
    [0, 0, 0, 1],
  ])
  view_control.convert_from_pinhole_camera_parameters(cam_params)

  iterations = range(len(verts))
  if verbose:
    iterations = progress_bar(iterations)
  for idx in iterations:
    mesh.vertices = o3d.utility.Vector3dVector(verts[idx])
    mesh.compute_vertex_normals()
    if colors is not None:
      mesh.vertex_colors = o3d.utility.Vector3dVector(colors[idx])
    vis.update_geometry(mesh)
    vis.poll_events()
    vis.update_renderer()
    frame = (np.asarray(vis.capture_screen_float_buffer()) * 255).astype(np.uint8)
    writer.write_frame(frame)

  writer.close()


def create_o3d_mesh(verts, faces, colors=None):
  """
  Create a open3d mesh from vertices and faces.

  Parameters
  ----------
  verts : np.ndarray, shape [v, 3]
    Mesh vertices.
  faces : np.ndarray, shape [f, 3]
    Mesh faces.

  Returns
  -------
  o3d.geometry.TriangleMesh
    Open3d mesh.
  """
  mesh = o3d.geometry.TriangleMesh()
  mesh.triangles = o3d.utility.Vector3iVector(faces)
  mesh.vertices = o3d.utility.Vector3dVector(verts)
  if colors is not None:
    mesh.vertex_colors = o3d.utility.Vector3dVector(colors)
  mesh.compute_vertex_normals()
  return mesh


def vis_mesh(verts, faces, width=1080, height=1080):
  """
  Visualize mesh with open3d.

  Parameters
  ----------
  verts : np.ndarray, shape [v, 3]
    Mesh vertices.
  faces : np.ndarray, shape [f, 3]
    Mesh faces.
  width : int
    Window width, by default 1080.
  height : int
    Window height, by default 1080.
  """
  mesh = create_o3d_mesh(verts, faces)
  viewer = o3d.visualization.Visualizer()
  viewer.create_window(width=width, height=height, visible=True)
  viewer.add_geometry(mesh)
  viewer.run()



import cv2
import numpy as np
import matplotlib.pyplot as plt
from .misc import *
from .io import *


def imshow_cv(img, caption='OpenCV Image Show'):
  """
  Show an image with opencv.

  Parameters
  ----------
  img : np.ndarray
    Input image in RGB format.
  caption : str, optional
    Window caption, by default 'OpenCV Image Show'
  """
  cv2.imshow(caption, np.flip(img, axis=-1).copy())
  cv2.waitKey()


def imshow(img, save_path=None):
  """
  Show an image with matplotlib.

  Parameters
  ----------
  img : np.ndarray, or list
    Image to show.
  save_path : str
    Path to save the figure.
  """
  if type(img) == np.ndarray:
    plt.imshow(img)
    if save_path is not None:
      plt.savefig(save_path)
    plt.show()
    plt.close()
  elif type(img) == list:
    imshow_grid(img)
  else:
    raise NotImplementedError('Unsupported type for visualization: ' + str(type(img)))


def imshow_grid(imgs, save_path=None):
  """
  Display multiple images as a grid.

  Parameters
  ----------
  imgs : list
    A list of images.
  nrows : int
    Number of rows.
  ncols : int
    Number of columns.
  save_path : str
    Path to save the figure.
  """
  fig = plt.figure()
  n_rows = len(imgs)
  n_cols = len(imgs[0])
  for i, img in enumerate(sum(imgs, [])):
    fig.add_subplot(n_rows, n_cols, i+1)
    plt.imshow(img)
  if save_path is not None:
    plt.savefig(save_path)
  plt.show()
  plt.close(fig)


def render_dots_from_uv(uv, canvas, id_label=False, radius=None):
  if canvas.dtype != np.uint8:
    print('canvas must be uint8 type')
    exit(0)
  if radius is None:
    radius = int(max(round(canvas.shape[0] / 128), 1))
  font_scale = canvas.shape[0] / 480
  uv = np.round(uv).astype(np.int32)
  for i in range(uv.shape[0]):
    cv2.circle(canvas, (uv[i][1], uv[i][0]), radius, (255, 0, 0), thickness=-1)
    if id_label:
      cv2.putText(
        canvas, '%d' % i, (uv[i][1], uv[i][0]), cv2.FONT_HERSHEY_PLAIN,
        font_scale, (0, 0, 255)
      )
  return canvas


def render_bones_from_uv(uv, canvas, skeleton, valid=None, colors=None,
                         thickness=None, save_path=None):
  """
  Render bones from joint uv coordinates.

  Parameters
  ----------
  uv : np.ndarray, shape [k, 2]
    UV coordinates of joints.
  canvas : np.ndarray, dtype uint8
    Canvas to draw on.
  colors : list
    Colors for each bone. A list of list in range [0, 255].
  thickness : int, optional
    Thickness of the line, by default None

  Returns
  -------
  np.ndarray
    The canvas after rendering.
  """
  if canvas.dtype != np.uint8:
    print('canvas must be uint8 type')
    exit(0)
  if thickness is None:
    thickness = int(max(round(canvas.shape[0] / 128), 1))

  if colors is None:
    if hasattr(skeleton, 'colors'):
      colors = skeleton.colors
    else:
      colors = [[255, 0, 0]] * len(skeleton.parents)

  for child, parent in enumerate(skeleton.parents):
    if parent is None:
      continue
    else:
      c = colors[child]
    start = (int(uv[parent][1]), int(uv[parent][0]))
    end = (int(uv[child][1]), int(uv[child][0]))

    if valid is not None:
      if not valid[parent] or not valid[child]:
        continue

    cv2.line(canvas, start, end, c, thickness)

  if save_path is not None:
    save(save_path, canvas)

  return canvas


def render_bones_from_hmap(hmap, canvas, skeleton, valid=None, colors=None,
                           thickness=None, save_path=None):
  """
  Render bones from heat maps.

  Parameters
  ----------
  hmap : np.ndarray, shape [h, w, k]
    Heat maps.
  parents : list
    Parent joint of each child joint.
  thickness : int, optional
    Thickness of each bone, by default None

  Returns
  -------
  np.ndarray
    Canvas after rendering.
  """
  coords = hmap_to_uv(hmap)
  bones = render_bones_from_uv(
    coords, canvas, skeleton, valid, colors, thickness, save_path
  )
  return bones


def render_bones_plt(joints, parents):
  """
  Render bones in 3D with matplotlib.

  Parameters
  ----------
  joints : np.ndarray
    Joint positions.
  parents : list
    Parent joint of each joint.
  """
  from mpl_toolkits.mplot3d import Axes3D
  fig = plt.figure()
  ax = Axes3D(fig)

  ax.set_xlim3d(-1.5, 1.5)
  ax.set_ylim3d(-1.5, 1.5)
  ax.set_zlim3d(-1.5, 1.5)

  ax.grid(False)
  ax.set_xticks([])
  ax.set_yticks([])
  ax.set_zticks([])
  ax.set_axis_off()
  ax.view_init(-90, -90)

  for c, p in enumerate(parents):
    if p is None:
      continue
    xs = [joints[c, 0], joints[p, 0]]
    ys = [joints[c, 1], joints[p, 1]]
    zs = [joints[c, 2], joints[p, 2]]
    plt.plot(xs, ys, zs, c=color_lib[p % len(color_lib)])
  plt.show()


def put_text(img, text, origin=None, color=(0, 255, 0), size=None):
  """
  origin : tuple, optional
    (x, y), by default None
  """
  font = cv2.FONT_HERSHEY_DUPLEX
  if size == None:
    size = int(math.ceil(max(img.shape) / 128))
  box = cv2.getTextSize(text, fontFace=font, fontScale=size, thickness=size)
  if origin is None:
    origin = (10, 10)
  origin = (origin[0], origin[1] + box[1])
  cv2.putText(img, text, origin, font, size, color=color, thickness=size)
  return img


def concat_videos(src_paths, tar_path, height=None, width=None):
  """
  You should only set height or width.
  """
  if height is None and width is None:
    raise RuntimeError('You must set either the height or the width.')
  readers = [VideoReader(p) for p in src_paths]
  writer = None
  for _ in progress_bar(readers[0].n_frames + 10):
    canvas = []
    for r in readers:
      frame = r.next_frame()
      if frame is None:
        canvas = []
        break
      canvas.append(imresize_diag(frame, width, height))
    if not canvas:
      break
    if height is None:
      canvas = np.concatenate(canvas, 0)
    else:
      canvas = np.concatenate(canvas, 1)
    if writer is None:
      writer = \
        VideoWriter(tar_path, canvas.shape[1], canvas.shape[0], readers[0].fps)
    writer.write_frame(canvas)
  for r in readers:
    r.close()
  writer.close()



import pickle
import numpy as np

from . import math_np


class LBSMesh():
  def __init__(self, model_path, skeleton, dtype=np.float32, fix_shape_bug=False):
    with open(model_path, 'rb') as f:
      data = pickle.load(f, encoding='latin1')

    if fix_shape_bug:
      data['shapedirs'] = np.array(data['shapedirs'])
      data['shapedirs'][:, 0, :] *= -1

    self.mesh = data['v_template'].astype(dtype)
    self.n_verts = self.mesh.shape[0]

    self.keypoints_mean = np.empty([skeleton.n_keypoints, 3], dtype)
    self.keypoints_mean[:data['J'].shape[0]] = data['J']
    for k, v in skeleton.extended_keypoints.items():
      self.keypoints_mean[k] = self.mesh[v]

    self.j_regressor = np.zeros([skeleton.n_keypoints, self.n_verts], dtype)

    # some early version uses scipy sparse array
    if type(data['J_regressor']) != np.ndarray:
      data['J_regressor'] = data['J_regressor'].toarray()

    self.j_regressor[:data['J_regressor'].shape[0]] = data['J_regressor']
    for k, v in skeleton.extended_keypoints.items():
      self.j_regressor[k] = 0
      self.j_regressor[k, v] = 1
    self.keypoints_std = np.einsum(
      'vdc, jv -> cjd', np.array(data['shapedirs'], dtype), self.j_regressor
    )

    self.parents = skeleton.parents
    self.children = [[] for _ in skeleton.parents]
    for c, p in enumerate(self.parents):
      if p is not None:
        self.children[p].append(c)

    # translate skinning weight: we use child joint
    self.skinning_weights = np.zeros(
      [data['weights'].shape[0], skeleton.n_keypoints], dtype=np.float32
    )
    for c, p in enumerate(self.parents):
      if p is not None:
        self.skinning_weights[:, c] = \
          data['weights'][:, p] / len(self.children[p])

    self.faces = data['f']
    self.shape_std = np.array(data['shapedirs'], dtype)
    self.ones = np.ones([1, self.n_verts, 1], dtype)
    self.skeleton = skeleton
    self.shape_dim = self.shape_std[-1]
    self.n_faces = self.faces.shape[0]
    self.dtype = dtype

  def pose_parent_to_children(self, pose, batch=False):
    if not batch:
      pose = np.expand_dims(pose, 0)

    # convert pose from children style to parent style
    outputs = [
      np.tile(np.expand_dims(np.eye(3), 0), [pose.shape[0], 1, 1]) \
        for _ in range(self.skeleton.n_keypoints)
    ]
    for c, p in enumerate(self.parents):
      if p is not None and p < pose.shape[1]:
        outputs[c] = pose[:, p]

    outputs = np.stack(outputs, 1)

    if not batch:
      outputs = outputs[0]

    return outputs

  def set_params(self, pose=None, shape=None, format='rotmat', relative=False,
                 reference='child', use_j_regressor=False, batch=False,
                 return_details=False, return_mesh=False):
    if pose is not None:
      pose = pose.astype(self.dtype)
    if shape is not None:
      shape = shape.astype(self.dtype)

    if not batch:
      if pose is not None:
        pose = np.expand_dims(pose, 0)
      if shape is not None:
        shape = np.expand_dims(shape, 0)

    verts = np.expand_dims(self.mesh.copy(), 0)

    keypoints = np.einsum('nvd, jv -> njd', verts, self.j_regressor)

    if shape is not None:
      verts = verts + np.einsum('nc, vdc -> nvd', shape.astype(self.dtype), self.shape_std)

    keypoints = np.einsum('nvd, jv -> njd', verts, self.j_regressor)

    if pose is None:
      if not batch:
        verts = verts[0]
        keypoints = keypoints[0]
      return keypoints, verts, None

    if format != 'rotmat':
      pose = math_np.convert(pose, format, 'rotmat')

    if reference == 'parent':
      pose = self.pose_parent_to_children(pose, batch=True)

    if relative:
      pose = math_np.rotmat_rel_to_abs(pose, self.parents, batch=True)

    bones = math_np.keypoints_to_bones(keypoints, self.parents, batch=True)
    posed_keypoints, _ = \
      math_np.forward_kinematics(bones, pose, self.parents, batch=True)

    if return_mesh:
      j_mat = posed_keypoints - np.einsum('njhw, njw -> njh', pose, keypoints)
      g_mat = np.concatenate([pose, np.expand_dims(j_mat, -1)], -1)
      verts = np.concatenate(
        [
          np.tile(verts, [pose.shape[0], 1, 1]),
          np.tile(self.ones, [pose.shape[0], 1, 1])
        ], -1
      )
      posed_verts = np.einsum(
        'vj, njvd -> nvd',
        self.skinning_weights, np.einsum('njhw, nvw -> njvh', g_mat, verts)
      )

    if use_j_regressor:
      posed_keypoints = \
        np.einsum('jv, nvd -> njd', self.j_regressor, posed_verts)

    if not batch:
      posed_keypoints = posed_keypoints[0]
      pose = pose[0]
      if return_mesh:
        posed_verts = posed_verts[0]

    if return_details:
      pack = {
        'posed_keypoints': posed_keypoints,
        'abs_rotmat': pose,
        'ref_bones': bones
      }
      if return_mesh:
        pack['posed_verts'] = posed_verts
      return pack

    if return_mesh:
      return posed_keypoints, posed_verts
    else:
      return posed_keypoints



import numpy as np
import cv2
import time
import uuid as uuid_import
import transforms3d
import tqdm
import os
import math

from .io import load_hdf5


color_lib = [
  [255, 0, 0], [230, 25, 75], [60, 180, 75], [255, 225, 25],
  [0, 130, 200], [245, 130, 48], [145, 30, 180], [70, 240, 240],
  [240, 50, 230], [210, 245, 60], [250, 190, 190], [0, 128, 128],
  [230, 190, 255], [170, 110, 40], [255, 250, 200], [128, 0, 0],
  [170, 255, 195], [128, 128, 0], [255, 215, 180], [0, 0, 128],
  [128, 128, 128]
]


def get_left_right_color(labels,
                         left=[255, 0, 0], right=[0, 255, 0], other=[0, 0, 255]):
  colors = []
  for l in labels:
    if 'left' in l:
      colors.append(left)
    elif 'right' in l:
      colors.append(right)
    else:
      colors.append(other)
  return colors


def imresize(img, size):
  """
  Resize an image.

  Parameters
  ----------
  img : np.ndarray
    Input image.
  size : tuple
    Size, (width, height).

  Returns
  -------
  np.ndarray
    Resized image.
  """
  return cv2.resize(img, size, cv2.INTER_LINEAR)


def imresize_diag(img, w=None, h=None):
  """
  Resize am image but keep the aspect ratio, according to target width or height.

  Parameters
  ----------
  img : [H, W, C]
    Image.
  w : int, optional
    Target width, by default None
  h : int, optional
    Target height, by default None

  Returns
  -------
  [H, W, C]
    Image.
  """
  if w is None:
    w = int(round(h / img.shape[0] * img.shape[1]))
  else:
    h = int(round(w / img.shape[1] * img.shape[0]))
  return imresize(img, (w, h))


def render_gaussian_hmap(centers, shape, sigma=None, dtype=np.float32):
  """
  Render gaussian heat maps from given centers.

  Parameters
  ----------
  centers : np.ndarray, shape [N, 2]
    Gussian centers, (row, column).
  shape : tuple
    Heatmap shape, (height, width)
  sigma : float, optional
    sigma, if None, would be height / 40

  Returns
  -------
  np.ndarray, float32, shape [height, width, N]
    Rendered heat maps
  """
  if sigma is None:
    sigma = shape[0] / 40
  x = [i for i in range(shape[1])]
  y = [i for i in range(shape[0])]
  xx, yy = np.meshgrid(x, y)
  xx = np.reshape(xx.astype(dtype), [shape[0], shape[1], 1])
  yy = np.reshape(yy.astype(dtype), [shape[0], shape[1], 1])
  x = np.reshape(centers[:,1], [1, 1, -1])
  y = np.reshape(centers[:,0], [1, 1, -1])
  distance = np.square(xx - x, dtype=dtype) + np.square(yy - y, dtype=dtype)
  hmap = np.exp(-distance / (2 * sigma**2 ), dtype=dtype) / \
         np.sqrt(2 * np.pi * sigma**2, dtype=dtype)
  hmap /= (np.max(hmap, axis=(0, 1), keepdims=True) + np.finfo(dtype).eps)
  return hmap


def print_important(s, sep='='):
  """
  Print important message.

  Parameters
  ----------
  s : str
    Message.
  sep : str, optional
    Seperator to fill one line, by default '='
  """
  print(sep*89)
  print(s)
  print(sep*89)


def arr_identical(a, b, verbose=True):
  """
  Wether two arrays are identical.

  Parameters
  ----------
  a : np.ndarray
    Array a.
  b : np.ndarray
    Array b.
  verbose : bool, optional
    Wether print detail infomation, by default True

  Returns
  -------
  bool
    Identical or not.
  """
  if a.shape != b.shape:
    if verbose:
      print('Different shape: a: {}, b: {}'.format(a.shape, b.shape))
    return False
  else:
    return np.allclose(a, b)


class Timer:
  """
  Timer to evaluate passed time.

  Call `tic` to save current time.

  """
  def __init__(self):
    self.last_tic = time.time()
    self.memory = {}

  def tic(self, key=None, reset=True):
    """
    Save current timestamp and return the interval from previous tic.

    Parameters
    ----------
    key : str, optional
      Key to save this interval, by default None
    reset : bool, optional
      If to reset the last tic time, by default True

    Returns
    -------
    float
      Interval time in second.
    """
    curr_time = time.time()
    interval = 0.0
    if self.last_tic is not None:
      interval = curr_time - self.last_tic
      if key is not None:
        self.memory[key] = interval
    if reset:
      self.last_tic = curr_time
    return interval


class LowPassFilter:
  """
  Lowpass filter. s = a * s + (1 - a) * s'

  Call 'process` to filter one signal.

  """
  def __init__(self, alpha=0.9):
    self.prev_raw_value = None
    self.prev_filtered_value = None
    self.alpha = alpha

  def process(self, value, alpha=None):
    """
    Filter the value.

    Parameters
    ----------
    value : np.ndarray
      Value to be filtered.

    Returns
    -------
    np.ndarray
      Filtered value.
    """
    if alpha is None:
      alpha = self.alpha
    if self.prev_raw_value is None:
      s = value
    else:
      s = alpha * value + (1.0 - alpha) * self.prev_filtered_value
    self.prev_raw_value = value
    self.prev_filtered_value = s
    return s


class OneEuroFilter:
  """
  One euro filter.

  Call `process` to process the signal.

  Parameters
  ----------
  mincutoff : float, optional
    Decrease mincutoff to decrease slow speed jittering, by default 1.0
  beta : float, optional
    Increase beta to decrease speed lag, by default 0.0
  dcutoff : float, optional
    by default 1.0
  freq : int, optional
    by default 30
  """
  def __init__(self, mincutoff=1.0, beta=0.0, dcutoff=1.0, freq=30):
    """
    Parameters
    ----------
    mincutoff : float, optional
      Decrease mincutoff to decrease slow speed jittering, by default 1.0
    beta : float, optional
      Increase beta to decrease speed lag, by default 0.0
    dcutoff : float, optional
      by default 1.0
    freq : int, optional
      by default 30
    """
    self.freq = freq
    self.mincutoff = mincutoff
    self.beta = beta
    self.dcutoff = dcutoff
    self.x_filter = LowPassFilter()
    self.dx_filter = LowPassFilter()

  def compute_alpha(self, cutoff):
    te = 1.0 / self.freq
    tau = 1.0 / (2 * np.pi * cutoff)
    return 1.0 / (1.0 + tau / te)

  def process(self, x):
    prev_x = self.x_filter.prev_raw_value
    dx = 0.0 if prev_x is None else (x - prev_x) * self.freq
    edx = self.dx_filter.process(dx, self.compute_alpha(self.dcutoff))
    cutoff = self.mincutoff + self.beta * np.abs(edx)
    return self.x_filter.process(x, self.compute_alpha(cutoff))


def hmap_to_uv(hmap):
  """
  Compute uv coordinates of heat map maxima in each layer.

  Parameters
  ----------
  hmap : np.ndarray, shape [h, w, k]
    Heat maps.

  Returns
  -------
  np.ndarray, shape [k, 2]
    UV coordinates in the order of (row, column).
  """
  shape = hmap.shape
  hmap = np.reshape(hmap, [-1, shape[-1]])
  v, h = np.unravel_index(np.argmax(hmap, 0), shape[:-1])
  coord = np.stack([v, h], 1)
  return coord


def camera_proj(k, xyz):
  """
  Camera projection: from camera space to image space.

  Parameters
  ----------
  k : np.ndarray, shape [3, 3]
    Camera intrinsics.
  xyz : np.ndarray, shape [n, 3]
    Coordinates.

  Returns
  -------
  np.ndarray, shape [n, 2]
    UV coordinates, (row, column)
  """
  uvd = np.dot(k, xyz.T)
  uv = np.flip((uvd / uvd[2:3, :])[:2].T, -1).copy()
  return uv


def camera_intrinsic(fx, fy, tx, ty):
  """
  Convert camera parameters into camera intrinsic matrix.

  Parameters
  ----------
  fx : float
    fx.
  fy : float
    fy
  tx : float
    tx
  ty : float
    ty

  Returns
  -------
  np.ndarray, shape [3, 3]
    Camear intrinsic matrix (placed on left).
  """
  return np.array([[fx, 0, tx], [0, fy, ty], [0, 0, 1]])


def compute_auc(xs, ys):
  """
  Compute area under curve (AuC).

  Parameters
  ----------
  xs : list
    A list of x values.
  ys : list
    A list of y values corresponding to x.

  Returns
  -------
  float
    Area under curve.
  """
  length = xs[-1] - xs[0]
  area = 0
  for i in range(len(ys) - 1):
    area += (ys[i] + ys[i + 1]) * (xs[i + 1] - xs[i]) / 2 / length
  return area


def compute_pck(errors, thres_range, n_step):
  """
  Compute percentage of correct keypoints (PCK) under a range of thresholds.

  Parameters
  ----------
  errors : np.ndarray, shape [n]
    Errors of all joints of all test samples.
  thres_range : tuple
    Threshold range, (min, max)
  n_step : int, optional
    Number of steps between the threshold range, by default 16

  Returns
  -------
  list
    Error thresholds.
  list
    PCK for each threshold.
  """
  xs = np.linspace(thres_range[0], thres_range[1], num=n_step)
  ys = []
  for x in xs:
    ys.append(np.sum(errors < x) / errors.shape[0])
  return xs, ys


def uuid():
  """
  Get a random string.

  Returns
  -------
  str
    Random string.
  """
  return uuid_import.uuid4()


def progress_bar(producer, text=None):
  if type(producer) == int:
    producer = range(producer)
  return tqdm.tqdm(list(producer), ascii=True, desc=text, dynamic_ncols=True)


def get_bbox(uv, scale, limit,
             rand_scale=False, rand_min=1.0, rand_max=2.0, square=False):
  """
  Get the bounding box of give uv coordinates.

  Parameters
  ----------
  uv : [N, 2]
    UV coordinates.
  scale : float
    Scale of the bounding box.
  limit : [2]
    The bottom-right of the bbox will be clipped to [0, limit]
  rand_scale : bool, optional
    Random scaling, by default False
  rand_min : float, optional
    Min random scale factor, by default 1.0
  rand_max : float, optional
    Max random scale factor, by default 2.0
  square : bool, optional
    Wether to make the bbox square, by default False
  """
  tl = np.min(uv, axis=0)
  br = np.max(uv, axis=0)
  center = (br + tl) / 2

  size = br - tl
  if square:
    size = np.max(size)
  if rand_scale:
    size *= np.random.uniform(rand_min, rand_max)
  else:
    size *= scale

  tl = np.round(np.clip(center - size / 2, 0, limit)).astype(np.int32)
  br = np.round(np.clip(center + size / 2, 0, limit)).astype(np.int32)

  return tl, br


def press_to_continue(exit_0=True):
  """
  Wait for user input to continue. Enter 'n' to exit.

  Parameters
  ----------
  exit_0 : bool, optional
    If True, will directly exit(0), instead of return False, by default True

  Returns
  -------
  bool
    Continue or not.
  """
  if input('Continue? (enter n to exit) ') == 'n':
    if exit_0:
      exit(0)
    else:
      return False
  return True


def inspect(data, indent=0):
  print(' ' * indent + f'Data type: {type(data)}')
  if type(data) == list:
    print(' ' * indent + f'length: {len(data)} first item:')
    inspect(data[0], indent=indent + 2)
  elif type(data) == dict:
    for k, v in data.items():
      print('  ' * indent, k, type(v))
      inspect(v, indent=indent + 2)
  else:
    if hasattr(data, 'shape'):
      print(' ' * indent + f'shape = {data.shape}')
    if hasattr(data, 'dtype'):
      print(' ' * indent + f'dtype = {data.dtype}')


examine_dict = inspect


def set_extension(file_name, ext):
  return os.path.splitext(file_name)[0] + '.' + ext


def count_model_parameters(model):
  return sum(p.numel() for p in model.parameters() if p.requires_grad)



import pygame
import numpy as np
import datetime

from transforms3d.axangles import axangle2mat
from OpenGL.GL import *
from OpenGL.GLU import *
from ..io import imsave, VideoWriter


class TriMeshViewer:
  """
  Render and visualize a triangle mesh with animation. See `TriMeshViewer.run`.

  Call `run()` to set up meshes and run.

  num-pad 5: reset view.
  space: play / pause.
  d: revert.
  f: forward.
  p: save screenshot.

  """
  def __init__(self, size=(512, 512)):
    self.rot_mat = np.eye(3)
    self.trans_v = np.zeros(3)
    self.speed_rot = np.pi / 45
    self.speed_trans = 0.02
    self.done = False
    self.playing = False
    self.window_size = size
    self.cam_offset = 1.5
    self.verts = []
    self.faces = np.empty([3, 3]) # just whatever
    self.n_verts = None
    self.n_faces = None
    self.frame_idx = None
    self.clock = pygame.time.Clock()

    self.dup_faces = None
    self.dup_normals = None
    self.verts_mapping = None
    self.normal_mapping = None

    self.n_frames = -1
    self.setup_opengl()

  def get_normal(self, faces):
    '''Given faces, return per face unit normal vector.'''
    p1x, p1y, p1z = faces[:, 0, 0], faces[:, 0, 1], faces[:, 0, 2]
    p2x, p2y, p2z = faces[:, 1, 0], faces[:, 1, 1], faces[:, 1, 2]
    p3x, p3y, p3z = faces[:, 2, 0], faces[:, 2, 1], faces[:, 2, 2]
    x_stick = (p2y - p1y) * (p3z - p1z) - (p2z - p1z) * (p3y - p1y)
    y_stick = (p2z - p1z) * (p3x - p1x) - (p2x - p1x) * (p3z - p1z)
    z_stick = (p2x - p1x) * (p3y - p1y) - (p2y - p1y) * (p3x - p1x)
    normals = np.stack((x_stick, y_stick, z_stick), axis=1)
    normals /=np.linalg.norm(normals, axis=1, keepdims=True)
    return normals

  def setup_dup(self):
    """
    For vertices normal computation.
    Since OpenGL only supports per vertex normal, here we duplicate the vertices
    so that each face has 3 same-normal-vertices belong to itself only.
    """
    self.dup_faces = np.empty([self.n_faces, 3], np.int32)
    self.verts_mapping = np.empty(self.n_faces * 3, np.int32)
    self.normal_mapping = np.empty(self.n_faces * 3, np.int32)

    for i in range(self.n_faces):
      for j in range(3):
        self.dup_faces[i, j] = i * 3 + j
        self.verts_mapping[i * 3 + j] = self.faces[i][j]
        self.normal_mapping[i * 3 + j] = i

  def setup_opengl(self):
    pygame.init()
    self.screen = \
      pygame.display.set_mode(self.window_size, pygame.DOUBLEBUF|pygame.OPENGL)
    pygame.display.set_caption('Frame %d / %d' % (0, self.n_frames))
    self.clock = pygame.time.Clock()

    glClearColor(1.0, 1.0, 1.0, 0)
    glShadeModel(GL_FLAT)

    glMaterialfv(GL_FRONT, GL_AMBIENT, np.array([.3]*3, dtype=np.float32))
    glMaterialfv(GL_FRONT, GL_DIFFUSE, np.array([1.0]*3, dtype=np.float32))
    glMaterialfv(GL_FRONT, GL_SPECULAR, np.array([.0]*3, dtype=np.float32))
    glMaterialf(GL_FRONT, GL_SHININESS,.4 * 128.0)

    glLightfv(GL_LIGHT0,GL_POSITION,np.array([0.0,-1.0,0.0,0.0],dtype=np.float32))
    glLightfv(GL_LIGHT0,GL_SPECULAR,np.array([0.0,0.0,0.0,1],dtype=np.float32))
    glLightfv(GL_LIGHT0,GL_DIFFUSE,np.array([1.0,1.0,1.0,1],dtype=np.float32))
    glLightfv(GL_LIGHT0,GL_AMBIENT,np.array([0.3,0.3,0.3,1],dtype=np.float32))
    glEnable(GL_LIGHT0)

    glLightfv(GL_LIGHT1,GL_POSITION,np.array([0.0,0.0,-1.0,0.0],dtype=np.float32))
    glLightfv(GL_LIGHT1,GL_SPECULAR,np.array([0.0,0.0,0.0,1],dtype=np.float32))
    glLightfv(GL_LIGHT1,GL_DIFFUSE,np.array([1.0,1.0,1.0,1],dtype=np.float32))
    glLightfv(GL_LIGHT1,GL_AMBIENT,np.array([0.3,0.3,0.3,1],dtype=np.float32))
    glEnable(GL_LIGHT1)

    glLightfv(GL_LIGHT2,GL_POSITION,np.array([-1.0,0.0,0.0,0.0],dtype=np.float32))
    glLightfv(GL_LIGHT2,GL_SPECULAR,np.array([0.0,0.0,0.0,1],dtype=np.float32))
    glLightfv(GL_LIGHT2,GL_DIFFUSE,np.array([1.0,1.0,1.0,1],dtype=np.float32))
    glLightfv(GL_LIGHT2,GL_AMBIENT,np.array([0.3,0.3,0.3,1],dtype=np.float32))
    glEnable(GL_LIGHT2)

    glEnable(GL_LIGHTING)
    glEnable(GL_DEPTH_TEST)
    glEnable(GL_VERTEX_ARRAY)
    glEnable(GL_NORMAL_ARRAY)
    gluPerspective(45, (self.window_size[0]/self.window_size[1]), 0.001, 10.0)

  def process_events(self):
    for event in pygame.event.get():
      if event.type == pygame.QUIT:
        self.done = True
      elif event.type == pygame.KEYDOWN:
        if event.key == pygame.K_KP5:
          self.trans_v = np.zeros(3)
          self.rot_mat = np.eye(3)
        elif event.key == pygame.K_SPACE:
          self.playing = not self.playing
        elif event.key == pygame.K_d:
          self.frame_idx -= 1
        elif event.key == pygame.K_f:
          self.frame_idx += 1
        elif event.key == pygame.K_p:
          img = pygame.image.tostring(self.screen, 'RGB')
          img = pygame.image.fromstring(img, self.window_size, 'RGB')
          img = np.transpose(pygame.surfarray.array3d(img), [1, 0, 2])
          imsave(
            '{date:%Y_%M_%d_%H_%M_%S}.png'.format(date=datetime.datetime.now()),
            img
          )
    pressed = pygame.key.get_pressed()

    # rotation
    if pressed[pygame.K_KP2]:
      rot_ax = [-1, 0, 0]
    elif pressed[pygame.K_KP8]:
      rot_ax = [1, 0, 0]
    elif pressed[pygame.K_KP4]:
      rot_ax = [0, -1, 0]
    elif pressed[pygame.K_KP6]:
      rot_ax = [0, 1, 0]
    elif pressed[pygame.K_KP1]:
      rot_ax = [0, 0, 1]
    elif pressed[pygame.K_KP3]:
      rot_ax = [0, 0, -1]
    else:
      rot_ax = None
    if rot_ax is not None:
      self.rot_mat = \
        np.matmul(axangle2mat(rot_ax, self.speed_rot), self.rot_mat)

    # translation
    if pressed[pygame.K_LEFT]:
      self.trans_v[0] += self.speed_trans
    if pressed[pygame.K_RIGHT]:
      self.trans_v[0] -= self.speed_trans
    if pressed[pygame.K_UP]:
      self.trans_v[1] -= self.speed_trans
    if pressed[pygame.K_DOWN]:
      self.trans_v[1] += self.speed_trans
    if pressed[pygame.K_q]:
      self.trans_v[2] += self.speed_trans
    if pressed[pygame.K_e]:
      self.trans_v[2] -= self.speed_trans

    if pressed[pygame.K_a]:
      self.frame_idx -= 1
    if pressed[pygame.K_s]:
      self.frame_idx += 1

    self.frame_idx = max(self.frame_idx, 0)
    self.frame_idx = min(self.frame_idx, self.n_frames)

  def norm_verts(self):
    maxi = np.max(self.verts, axis=(0, 1))
    mini = np.min(self.verts, axis=(0, 1))
    scale = np.max(maxi - mini)
    mean = np.mean(self.verts)
    self.verts = (self.verts - mean) / scale

  def render(self):
    verts = np.matmul(self.rot_mat, self.verts[self.frame_idx].T).T
    verts -= self.trans_v
    verts[..., 2] -= self.cam_offset

    face_coords = verts[self.faces]
    face_normals = self.get_normal(face_coords)

    dup_verts = verts[self.verts_mapping].copy()
    dup_normals = face_normals[self.normal_mapping].copy()

    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)
    glVertexPointerf(dup_verts)
    glNormalPointerf(dup_normals)
    glDrawElementsui(GL_TRIANGLES, self.dup_faces)

  def run(self, verts, faces, video_path=None, video_fps=30, run_fps=30):
    """
    Render mesh animation.

    Parameters
    ----------
    verts : np.ndarray, [n, v, 3]
      Mesh vertices.
    faces : np.ndarray, [f, 3]
      Mesh faces.
    video_path : str, optional
      Path to save the rendered video if not None, by default None
    video_fps : int, optional
      Frame rate of saved video., by default 30
    run_fps : int, optional
      Frame rate of renderer, by default 30
    """
    self.faces = faces.astype(np.int32).copy()
    if type(verts) == list:
      verts = np.stack(verts, 0)
    self.verts = verts.copy()
    self.n_frames, self.n_verts, _ = self.verts.shape
    self.n_faces = self.faces.shape[0]
    self.norm_verts()
    self.setup_dup()

    self.frame_idx = -1

    if video_path is not None:
      video = VideoWriter(
        video_path, self.window_size[0], self.window_size[1], video_fps
      )

    if video_path is not None:
      self.playing = True

    while not self.done:
      self.process_events()

      if self.frame_idx >= self.n_frames:
        if video_path is not None:
          self.done = True
        self.frame_idx = self.n_frames - 1
      pygame.display.set_caption(
        'Frame %d / %d' % (self.frame_idx + 1, self.n_frames)
      )

      self.render()

      if video_path is not None:
        frame = pygame.image.tostring(self.screen, 'RGB')
        frame = pygame.image.fromstring(frame, self.window_size, 'RGB')
        frame = np.transpose(pygame.surfarray.array3d(frame), [1, 0, 2])
        frame = np.flip(frame, -1).copy()
        video.write_frame(frame)

      pygame.display.flip()

      if self.playing:
        self.frame_idx += 1
      if self.frame_idx >= self.n_frames and video_path is not None:
        self.done = True

      if run_fps is not None and video_path is None:
        self.clock.tick(run_fps)

    pygame.quit()
    if video_path is not None:
      video.close()



import setuptools

setuptools.setup(
  name='vctoolkit',
  version='0.1.9.38',
  author_email='yuxiao.zhou@outlook.com',
  description='A simple wrapper for commonly used tools in visual computing.',
  url='https://github.com/CalciferZh/vctoolkit',
  packages=setuptools.find_packages(),
  classifiers=["Programming Language :: Python :: 3"],
  install_requires=[
    'opencv-python', 'h5py', 'imageio', 'numpy', 'scipy',
    'transforms3d', 'tqdm', 'matplotlib', 'pyyaml'
  ]
)



import numpy as np


def get_finger_color(labels):
  color = []
  for l in labels:
    if 'T' in l:
      c = [255, 0, 0]
    elif 'I' in l:
      c = [0, 255, 0]
    elif 'M' in l:
      c = [0, 0, 255]
    elif 'R' in l:
      c = [0, 255, 255]
    elif 'L' in l:
      c = [255, 0, 255]
    else:
      c = [255, 255, 0]
    color.append(c)
  return color


def get_body_color(labels):
  color = []
  for l in labels:
    if 'left' in l:
      c = [255, 0, 0]
    elif 'right' in l:
      c = [0, 255, 0]
    else:
      c = [0, 0, 255]
    color.append(c)
  return color


def convert_skeleton(data, src_skeleton, tgt_skeleton, axis=0):
  src_data = np.swapaxes(data, 0, axis)
  tgt_data = []
  for l in tgt_skeleton.labels:
    tgt_data.append(src_data[src_skeleton.labels.index(l)])
  tgt_data = np.swapaxes(np.stack(tgt_data), 0, axis).copy()
  return tgt_data


class BaseSkeleton:
  labels = []

  parents = [] # parent keypoint in the kinematic tree

  n_keypoints = 0

  center = None # for alignment

  root = None # the root joint in the kinematics tree


class COCOBody(BaseSkeleton):
  n_keypoints = 17

  labels = [
    "nose", # 0
    "left_eye", "right_eye", # 2
    "left_ear", "right_ear", # 4
    "left_shoulder", "right_shoulder", # 6
    "left_elbow", "right_elbow", # 8
    "left_wrist", "right_wrist", # 10
    "left_hip", "right_hip", # 12
    "left_knee", "right_knee", # 14
    "left_ankle", "right_ankle" # 16
  ]

  parents = [
    None,
    0, 0,
    1, 2,
    None, None,
    5, 6,
    7, 8,
    None, None,
    11, 12,
    13, 14
  ]

  center = root = None

  colors = get_body_color(labels)


class MPII2DBody(BaseSkeleton):
  n_keypoints = 16

  labels = [
    'right_ankle', 'right_knee', 'right_hip', # 2
    'left_hip', 'left_knee', 'left_ankle', # 5
    'pelvis', 'thorax', 'upper_neck', 'head_top', # 9
    'right_wrist', 'right_elbow', 'right_shoulder', # 12
    'left_shoulder', 'left_elbow', 'left_wrist' # 15
  ]

  parents = [
    1, 2, 6,
    6, 3, 4,
    None, 6, 7, 8,
    11, 12, 7,
    7, 13, 14
  ]

  center = root = 6

  colors = get_body_color(labels)


class MPII3DBody28(BaseSkeleton):
  n_keypoints = 28

  labels = [
    'spine3', 'spine4', 'spine2', 'spine', 'pelvis', # 4
    'neck', 'head', 'head_top', # 7
    'left_clavical', 'left_shoulder', 'left_elbow', 'left_wrist', 'left_hand', # 12
    'right_clavical', 'right_shoulder', 'right_elbow', 'right_wrist', 'right_hand', # 17
    'left_hip', 'left_knee', 'left_ankle', 'left_foot', 'left_toe', # 22
    'right_hip', 'right_knee', 'right_ankle', 'right_foot', 'right_toe', # 27
  ]

  parents = [
    2, 0, 3, 4,
    None, 1, 5, 6,
    1, 8, 9, 10, 11,
    1, 13, 14, 15, 16,
    4, 18, 19, 20, 21,
    4, 23, 24, 25, 26
  ]

  center = root = 4

  colors = get_body_color(labels)


class MPII3DBody17(BaseSkeleton):
  n_keypoints = 17

  labels = [
    'head_top', 'neck', # 1
    'right_shoulder', 'right_elbow', 'right_wrist', # 4
    'left_shoulder', 'left_elbow', 'left_wrist', # 7
    'right_hip', 'right_knee', 'right_ankle', # 10
    'left_hip', 'left_knee', 'left_ankle', # 13
    'pelvis', 'spine', 'head' # 16
  ]

  parents = [
    16, 15,
    1, 2, 3,
    1, 5, 6,
    14, 8, 9,
    14, 11, 12,
    None, 14, 1
  ]

  center = root = 14

  colors = get_body_color(labels)


class HM36MBody23(BaseSkeleton):
  labels = [
    'pelvis', 'spine', 'neck', 'head', 'head_top', # 4
    'left_shoulder', 'left_elbow', 'left_wrist', 'left_hand', # 8
    'right_shoulder', 'right_elbow', 'right_wrist', 'right_hand', # 12
    'left_hip', 'left_knee', 'left_ankle', 'left_foot', 'left_toe', # 17
    'right_hip', 'right_knee', 'right_ankle', 'right_foot', 'right_toe' # 22
  ]

  n_keypoints = 23

  parents = [
    None, 0, 1, 2, 3,
    2, 5, 6, 7,
    2, 9, 10, 11,
    0, 13, 14, 15, 16,
    0, 18, 19, 20, 21
  ]

  center = root = 0

  colors = get_body_color(labels)


class HM36MBody17(BaseSkeleton):
  labels = [
    'pelvis', 'spine', 'neck', 'head', 'head_top', # 4
    'left_shoulder', 'left_elbow', 'left_wrist', # 7
    'right_shoulder', 'right_elbow', 'right_wrist', # 10
    'left_hip', 'left_knee', 'left_ankle',  # 13
    'right_hip', 'right_knee', 'right_ankle', # 16
  ]

  n_keypoints = 17

  parents = [
    None, 0, 1, 2, 3,
    2, 5, 6,
    2, 8, 9,
    0, 11, 12,
    0, 14, 15
  ]

  center = root = 0

  colors = get_body_color(labels)


class HM36MBody32(BaseSkeleton):
  labels = [
    'pelvis', # 0
    'right_hip', 'right_knee', 'right_ankle', 'right_foot', 'right_toe', # 5
    'left_hip', 'left_knee', 'left_ankle', 'left_foot', 'left_toe', # 10
    'pelvis_duplicate', 'spine', 'neck', 'head', 'head_top', # 15
    'unknown', # 16
    'left_shoulder', 'left_elbow', 'left_wrist', # 19
    'unknown', 'unknown', # 21
    'left_hand', # 22
    'unknown', 'unknown', # 24
    'right_shoulder', 'right_elbow', 'right_wrist', # 27
    'unknown', 'unknown', # 29
    'right_hand', # 30
    'unknown', 'unknown', # 32
  ]

  n_keypoints = 32

  parents = [
    None,
    0, 1, 2, 3, 4,
    0, 6, 7, 8, 9,
    None, 0, 12, 13, 14,
    None,
    12, 17, 18,
    None, None,
    19,
    None, None,
    12, 25, 26,
    None, None,
    27,
    None, None
  ]

  center = root = 0

  colors = get_body_color(labels)


class HUMBIBody33(BaseSkeleton):
  n_keypoints = 33

  labels = [
    'pelvis', # 0
    'left_hip', 'right_hip', # 2
    'lowerback', # 3
    'left_knee', 'right_knee', # 5
    'upperback', # 6
    'left_ankle', 'right_ankle', # 8
    'thorax', # 9
    'left_toes', 'right_toes', # 11
    'lowerneck', # 12
    'left_clavicle', 'right_clavicle', # 14
    'upperneck', # 15
    'left_shoulder', 'right_shoulder', # 17
    'left_elbow', 'right_elbow', # 19
    'left_wrist', 'right_wrist', # 21
    # the fake hand joints in SMPL are removed
    # following are extended keypoints
    'head_top', 'left_eye', 'right_eye', # 24
    'left_hand_I0', 'left_hand_L0', # 26
    'right_hand_I0', 'right_hand_L0', # 28
    'left_foot_T0', 'left_foot_L0', # 30
    'right_foot_T0', 'right_foot_L0', # 32
  ]

  parents = [
    None,
    0, 0,
    0,
    1, 2,
    3,
    4, 5,
    6,
    7, 8,
    9,
    9, 9,
    12,
    13, 14,
    16, 17,
    18, 19,
    # extended
    15, 15, 15,
    20, 20,
    21, 21,
    7, 7,
    8, 8
  ]

  center = root = 0

  # the vertex indices of the extended keypoints
  extended_keypoints = {
    22: 411, 23: 2800, 24: 6260,
    25: 2135, 26: 2062,
    27: 5595, 28: 5525,
    29: 3292, 30: 3318,
    31: 6691, 32: 6718
  }

  colors = get_body_color(labels)


class MTCBody(BaseSkeleton):
  n_keypoints = 19

  labels = [
    'neck', 'nose', 'pelvis', # 2
    'left_shoulder', 'left_elbow', 'left_wrist', # 5
    'left_hip', 'left_knee', 'left_ankle', # 8
    'right_shoulder', 'right_elbow', 'right_wrist', # 11
    'right_hip', 'right_knee', 'right_ankle', # 14
    'left_eye', 'left_ear', # 16
    'right_eye', 'right_ear' # 18
  ]

  parents = [
    2, 0, None,
    0, 3, 4,
    2, 6, 7,
    0, 9, 10,
    2, 12, 13,
    1, 15,
    1, 17
  ]

  center = root = 2

  colors = get_body_color(labels)


class MTCWholeBody(BaseSkeleton):
  n_keypoints = 61

  labels = [
    'neck', 'nose', 'pelvis', # 2
    'left_shoulder', 'left_elbow', 'left_wrist', # 5
    'left_hip', 'left_knee', 'left_ankle', # 8
    'right_shoulder', 'right_elbow', 'right_wrist', # 11
    'right_hip', 'right_knee', 'right_ankle', # 14
    'left_eye', 'left_ear', # 16
    'right_eye', 'right_ear', # 18
    'left_W', # 19
    'left_T0', 'left_T1', 'left_T2', 'left_T3', # 23
    'left_I0', 'left_I1', 'left_I2', 'left_I3',
    'left_M0', 'left_M1', 'left_M2', 'left_M3',
    'left_R0', 'left_R1', 'left_R2', 'left_R3',
    'left_L0', 'left_L1', 'left_L2', 'left_L3', # 39
    'right_W', # 40
    'right_T0', 'right_T1', 'right_T2', 'right_T3', # 44
    'right_I0', 'right_I1', 'right_I2', 'right_I3',
    'right_M0', 'right_M1', 'right_M2', 'right_M3',
    'right_R0', 'right_R1', 'right_R2', 'right_R3',
    'right_L0', 'right_L1', 'right_L2', 'right_L3', # 60
  ]

  parents = [
    2, 0, None,
    0, 3, 4,
    2, 6, 7,
    0, 9, 10,
    2, 12, 13,
    1, 15,
    1, 17,
    4,
    19, 20, 21, 22,
    19, 24, 25, 26,
    19, 28, 29, 30,
    19, 32, 33, 34,
    19, 36, 37, 38,
    10,
    40, 41, 42, 43,
    40, 45, 46, 47,
    40, 49, 50, 51,
    40, 53, 54, 55,
    40, 57, 58, 59
  ]

  center = root = 2

  colors = get_body_color(labels)


class COCOWholeBody(BaseSkeleton):
  n_keypoints = 59

  labels = [
    "nose", # 0
    "left_eye", "right_eye", # 2
    "left_ear", "right_ear", # 4
    "left_shoulder", "right_shoulder", # 6
    "left_elbow", "right_elbow", # 8
    "left_wrist", "right_wrist", # 10
    "left_hip", "right_hip", # 12
    "left_knee", "right_knee", # 14
    "left_ankle", "right_ankle", # 16
    'left_W', # 17
    'left_T0', 'left_T1', 'left_T2', 'left_T3', # 21
    'left_I0', 'left_I1', 'left_I2', 'left_I3',
    'left_M0', 'left_M1', 'left_M2', 'left_M3',
    'left_R0', 'left_R1', 'left_R2', 'left_R3',
    'left_L0', 'left_L1', 'left_L2', 'left_L3', # 37
    'right_W', # 38
    'right_T0', 'right_T1', 'right_T2', 'right_T3', # 42
    'right_I0', 'right_I1', 'right_I2', 'right_I3',
    'right_M0', 'right_M1', 'right_M2', 'right_M3',
    'right_R0', 'right_R1', 'right_R2', 'right_R3',
    'right_L0', 'right_L1', 'right_L2', 'right_L3', # 58
  ]

  parents = [
    None,
    0, 0,
    1, 2,
    None, None,
    5, 6,
    7, 8,
    None, None,
    11, 12,
    13, 14,
    7,
    17, 18, 19, 20,
    17, 22, 23, 24,
    17, 26, 27, 28,
    17, 30, 31, 32,
    17, 34, 35, 36,
    8,
    38, 39, 40, 41,
    38, 43, 44, 45,
    38, 47, 48, 49,
    38, 51, 52, 53,
    38, 55, 56, 57,
  ]

  center = root = 0

  colors = get_body_color(labels)


class MANOHand(BaseSkeleton):
  n_keypoints = 21

  center = 4

  root = 0

  labels = [
    'W', #0
    'I0', 'I1', 'I2', #3
    'M0', 'M1', 'M2', #6
    'L0', 'L1', 'L2', #9
    'R0', 'R1', 'R2', #12
    'T0', 'T1', 'T2', #15
    'I3', 'M3', 'L3', 'R3', 'T3' #20, tips are manually added (not in MANO)
  ]

  # finger tips are not keypoints in MANO, we label them on the mesh manually
  extended_keypoints = {16: 333, 17: 444, 18: 672, 19: 555, 20: 744}

  parents = [
    None,
    0, 1, 2,
    0, 4, 5,
    0, 7, 8,
    0, 10, 11,
    0, 13, 14,
    3, 6, 9, 12, 15
  ]

  colors = get_finger_color(labels)


class MPIIHand(BaseSkeleton):
  n_keypoints = 21

  center = 9

  root = 0

  labels = [
    'W', #0
    'T0', 'T1', 'T2', 'T3', #4
    'I0', 'I1', 'I2', 'I3', #8
    'M0', 'M1', 'M2', 'M3', #12
    'R0', 'R1', 'R2', 'R3', #16
    'L0', 'L1', 'L2', 'L3', #20
  ]

  parents = [
    None,
    0, 1, 2, 3,
    0, 5, 6, 7,
    0, 9, 10, 11,
    0, 13, 14, 15,
    0, 17, 18, 19
  ]

  colors = get_finger_color(labels)


class RHDHand(BaseSkeleton):
  n_keypoints = 21

  center = 9

  root = 0

  labels = [
    'W', #0
    'T3', 'T2', 'T1', 'T0', #4
    'I3', 'I2', 'I1', 'I0', #4
    'M3', 'M2', 'M1', 'M0', #4
    'R3', 'R2', 'R1', 'R0', #4
    'L3', 'L2', 'L1', 'L0', #4
  ]

  parents = [
    None,
    2, 3, 4, 0, # 4
    6, 7, 8, 0, # 8
    10, 11, 12, 0, # 12
    14, 15, 16, 0, # 16
    18, 19, 20, 0, # 20
  ]

  colors = get_finger_color(labels)


class SMPLH65(BaseSkeleton):
  n_keypoints = 65

  center = root = 0

  labels = [
    'pelvis', # 0
    'left_hip', 'right_hip', # 2
    'lowerback', # 3
    'left_knee', 'right_knee', # 5
    'upperback', # 6
    'left_ankle', 'right_ankle', # 8
    'thorax', # 9
    'left_toes', 'right_toes', # 11
    'lowerneck', # 12
    'left_clavicle', 'right_clavicle', # 14
    'upperneck', # 15
    'left_shoulder', 'right_shoulder', # 17
    'left_elbow', 'right_elbow', # 19
    'left_wrist', 'right_wrist', # 21
    # left hand
    'LI0', 'LI1', 'LI2', # 24
    'LM0', 'LM1', 'LM2', # 27
    'LL0', 'LL1', 'LL2', # 30
    'LR0', 'LR1', 'LR2', # 33
    'LT0', 'LT1', 'LT2', # 36
    # right hand
    'RI0', 'RI1', 'RI2', # 39
    'RM0', 'RM1', 'RM2', # 42
    'RL0', 'RL1', 'RL2', # 45
    'RR0', 'RR1', 'RR2', # 48
    'RT0', 'RT1', 'RT2', # 51
    # extended keypoints
    # hand tips
    'LI3', 'LM3', 'LL3', 'LR3', 'LT3', # 56
    'RI3', 'RM3', 'RL3', 'RR3', 'RT3', # 61
    # foot middle toe root
    'LMT', 'RMT' # 63
    # head top
    'head_top' # 64
  ]

  parents = [
    None,
    0, 0,
    0,
    1, 2,
    3,
    4, 5,
    6,
    7, 8,
    9,
    9, 9,
    12,
    13, 14,
    16, 17,
    18, 19,
    # left hand
    20, 22, 23,
    20, 25, 26,
    20, 28, 29,
    20, 31, 32,
    20, 34, 35,
    # right hand
    21, 37, 38,
    21, 40, 41,
    21, 43, 44,
    21, 46, 47,
    21, 49, 50,
    # extended keypoints
    24, 27, 30, 33, 36,
    39, 42, 45, 48, 51,
    10, 11,
    15
  ]

  # the vertex indices of the extended keypoints
  extended_keypoints = {
    52: 2319, 53: 2445, 54: 2673, 55: 2556, 56: 2746,
    57: 5782, 58: 5905, 59: 6133, 60: 6016, 61: 6206,
    62: 3255, 63: 6655, 64: 411
  }

  colors = get_body_color(labels)


class SMPLBody22(BaseSkeleton):
  n_keypoints = 22

  center = root = 0

  labels = [
    'pelvis', # 0
    'left_hip', 'right_hip', # 2
    'lowerback', # 3
    'left_knee', 'right_knee', # 5
    'upperback', # 6
    'left_ankle', 'right_ankle', # 8
    'thorax', # 9
    'left_toes', 'right_toes', # 11
    'lowerneck', # 12
    'left_clavicle', 'right_clavicle', # 14
    'upperneck', # 15
    'left_shoulder', 'right_shoulder', # 17
    'left_elbow', 'right_elbow', # 19
    'left_wrist', 'right_wrist', # 21
  ]

  parents = [
    None,
    0, 0,
    0,
    1, 2,
    3,
    4, 5,
    6,
    7, 8,
    9,
    9, 9,
    12,
    13, 14,
    16, 17,
    18, 19,
  ]

  colors = get_body_color(labels)


class SMPLBody26(BaseSkeleton):
  n_keypoints = 26

  center = root = 0

  labels = [
    'pelvis', # 0
    'left_hip', 'right_hip', # 2
    'lowerback', # 3
    'left_knee', 'right_knee', # 5
    'upperback', # 6
    'left_ankle', 'right_ankle', # 8
    'thorax', # 9
    'left_toes', 'right_toes', # 11
    'lowerneck', # 12
    'left_clavicle', 'right_clavicle', # 14
    'upperneck', # 15
    'left_shoulder', 'right_shoulder', # 17
    'left_elbow', 'right_elbow', # 19
    'left_wrist', 'right_wrist', # 21
    'left_hand', 'right_hand', # 23
    'left_hand_tip', 'right_hand_tip' # 25
  ]

  parents = [
    None,
    0, 0,
    0,
    1, 2,
    3,
    4, 5,
    6,
    7, 8,
    9,
    9, 9,
    12,
    13, 14,
    16, 17,
    18, 19,
    20, 21,
    22, 23
  ]

  # the vertex indices of the extended keypoints
  extended_keypoints = {
    24: 2445, 25: 5905
  }

  colors = get_body_color(labels)


class SMPLBody35(BaseSkeleton):
  n_keypoints = 35

  labels = [
    'pelvis', # 0
    'left_hip', 'right_hip', # 2
    'lowerback', # 3
    'left_knee', 'right_knee', # 5
    'upperback', # 6
    'left_ankle', 'right_ankle', # 8
    'thorax', # 9
    'left_toes', 'right_toes', # 11
    'lowerneck', # 12
    'left_clavicle', 'right_clavicle', # 14
    'upperneck', # 15
    'left_shoulder', 'right_shoulder', # 17
    'left_elbow', 'right_elbow', # 19
    'left_wrist', 'right_wrist', # 21
    'left_hand', 'right_hand', # 23
    # the fake hand joints in SMPL are removed
    # following are extended keypoints
    'head_top', 'left_eye', 'right_eye', # 26
    'left_hand_I0', 'left_hand_L0', # 28
    'right_hand_I0', 'right_hand_L0', # 30
    'left_foot_T0', 'left_foot_L0', # 32
    'right_foot_T0', 'right_foot_L0', # 34
  ]

  parents = [
    None,
    0, 0,
    0,
    1, 2,
    3,
    4, 5,
    6,
    7, 8,
    9,
    9, 9,
    12,
    13, 14,
    16, 17,
    18, 19,
    20, 21,
    # extended
    15, 15, 15,
    22, 22,
    23, 23,
    10, 10,
    11, 11
  ]

  center = root = 0

  # the vertex indices of the extended keypoints
  extended_keypoints = {
    24: 411, 25: 2800, 26: 6260,
    27: 2135, 28: 2062,
    29: 5595, 30: 5525,
    31: 3292, 32: 3318,
    33: 6691, 34: 6718
  }

  colors = get_body_color(labels)


class LSPBody(BaseSkeleton):
  n_keypoints = 14

  labels = [
    'right_ankle', 'right_knee', 'right_hip', # 2
    'left_hip', 'left_knee', 'left_ankle', # 5
    'right_wrist', 'right_elbow', 'right_shoulder', # 8
    'left_shoulder', 'left_elbow', 'left_wrist', # 11
    'neck', 'head_top' # 13
  ]

  parents = [
    1, 2, None,
    None, 3, 4,
    7, 8, 12,
    12, 9, 10,
    None, 12
  ]

  center = root = 12

  colors = get_body_color(labels)



import numpy as np
from .io import *
from tqdm import tqdm
from transforms3d.axangles import axangle2mat


def joints_to_mesh(joints, skeleton, colors=None, thickness=0.2, save_path=None, valid=None):
  """
  Produce a mesh representing the skeleton with given joint coordinates.
  Bones are represented by prisms.

  Parameters
  ----------
  joints : np.ndarray, shape [k, 3]
    Joint coordinates.
  colors : list
    Colors for each bone. A list of list in range [0, 255].
  thickness : float, optional
    The thickness of the bone relative to length, by default 0.2

  Returns
  -------
  np.ndarray, shape [v, 3]
    Vertices of the mesh.
  np.ndarray, shape [f, 3]
    Face indices of the mesh.
  """
  if valid is None:
    valid = np.ones(joints.shape[0], np.bool)
  n_bones = len(list(filter(lambda x: x is not None, skeleton.parents)))
  # since some keypoints are invalid, the rendered bones will be less
  # this is the maximum buffer
  faces = np.empty([n_bones * 8, 3], dtype=np.int32)
  verts = np.empty([n_bones * 6, 3], dtype=np.float32)
  face_color = []
  bone_idx = -1
  for child, parent in enumerate(skeleton.parents):
    if parent is None:
      continue
    if not (valid[child] and valid[parent]):
      continue
    bone_idx += 1
    a = joints[parent]
    b = joints[child]
    ab = b - a
    f = a + thickness * ab

    if ab[0] == 0:
      ax = [0, 1, 0]
    else:
      ax = [-ab[1]/ab[0], 1, 0]

    fd = np.transpose(axangle2mat(ax, -np.pi/2).dot(np.transpose(ab))) \
         * thickness / 1.2
    d = fd + f
    c = np.transpose(axangle2mat(ab, -np.pi/2).dot(np.transpose(fd))) + f
    e = np.transpose(axangle2mat(ab, np.pi/2).dot(np.transpose(fd))) + f
    g = np.transpose(axangle2mat(ab, np.pi).dot(np.transpose(fd))) + f

    verts[bone_idx*6+0] = a
    verts[bone_idx*6+1] = b
    verts[bone_idx*6+2] = c
    verts[bone_idx*6+3] = d
    verts[bone_idx*6+4] = e
    verts[bone_idx*6+5] = g

    faces[bone_idx*8+0] = \
      np.flip(np.array([0, 2, 3], dtype=np.int32), axis=0) + bone_idx * 6
    faces[bone_idx*8+1] = \
      np.flip(np.array([0, 3, 4], dtype=np.int32), axis=0) + bone_idx * 6
    faces[bone_idx*8+2] = \
      np.flip(np.array([0, 4, 5], dtype=np.int32), axis=0) + bone_idx * 6
    faces[bone_idx*8+3] = \
      np.flip(np.array([0, 5, 2], dtype=np.int32), axis=0) + bone_idx * 6
    faces[bone_idx*8+4] = \
      np.flip(np.array([1, 4, 3], dtype=np.int32), axis=0) + bone_idx * 6
    faces[bone_idx*8+5] = \
      np.flip(np.array([1, 3, 2], dtype=np.int32), axis=0) + bone_idx * 6
    faces[bone_idx*8+6] = \
      np.flip(np.array([1, 5, 4], dtype=np.int32), axis=0) + bone_idx * 6
    faces[bone_idx*8+7] = \
      np.flip(np.array([1, 2, 5], dtype=np.int32), axis=0) + bone_idx * 6

    if colors is not None:
      for _ in range(8):
        face_color.append(colors[child])

  # since some keypoints are invalid, the rendered bones will be less
  faces = faces[:(bone_idx + 1)*8]
  verts = verts[:(bone_idx + 1)*6]

  if save_path is not None:
    save(save_path, (verts, faces))

  if colors is not None:
    return verts, faces, face_color

  return verts, faces



import numpy as np


def convert(rot, src, tar):
  # https://www.euclideanspace.com/maths/geometry/rotations/conversions/
  eps = np.finfo(np.float32).eps
  if src == 'axangle':
    data_shape = rot.shape[:-1]
    rot = np.reshape(rot, [-1, 3])
    if tar == 'quat':
      rad = np.linalg.norm(rot, axis=-1, keepdims=True)
      ax = rot / np.maximum(rad, eps)
      w = np.cos(rad / 2)
      xyz = np.sin(rad / 2) * ax
      quat = np.concatenate([w, xyz], -1)
      quat = np.reshape(quat, data_shape + (4,))
      return quat

    if tar == 'rotmat' or tar == 'rot6d':
      theta = np.linalg.norm(rot, axis=-1, keepdims=True)
      c = np.cos(theta)
      s = np.sin(theta)
      t = 1 - c
      x, y, z = np.split(rot / np.maximum(theta, eps), 3, axis=-1)
      rotmat = np.stack([
        t*x*x + c, t*x*y - z*s, t*x*z + y*s,
        t*x*y + z*s, t*y*y + c, t*y*z - x*s,
        t*x*z - y*s, t*y*z + x*s, t*z*z + c
      ], 1)
      rotmat = np.reshape(rotmat, data_shape + (3, 3))
      if tar == 'rotmat':
        return rotmat
      if tar == 'rot6d':
        return convert(rotmat, 'rotmat', 'rot6d')

  if src == 'quat':
    data_shape = rot.shape[:-1]
    rot = np.reshape(rot, [-1, 4])
    if tar == 'rotmat' or tar == 'rot6d':
      w, x, y, z = np.split(rot, 4, axis=-1)
      rotmat = np.stack([
        1-2*y*y-2*z*z, 2*x*y-2*z*w, 2*x*z+2*y*w,
        2*x*y+2*z*w, 1-2*x*x-2*z*z, 2*y*z-2*x*w,
        2*x*z-2*y*w, 2*y*z+2*x*w, 1-2*x*x-2*y*y
      ], -1)
      rotmat = np.reshape(rotmat, data_shape + (3, 3))
      if tar == 'rotmat':
        return rotmat
      if tar == 'rot6d':
        return convert(rotmat, 'rotmat', 'rot6d')

    if tar == 'axangle':
      angle = 2 * np.arccos(rot[:, 0:1])
      axis = rot[:, 1:] / np.sqrt(1 - np.square(rot[:, 0]))
      axangle = axis * angle
      axangle = np.reshape(axangle, data_shape + (3,))
      return axangle

  if src == 'rotmat':
    data_shape = rot.shape[:-2]
    rot = np.reshape(rot, [-1, 3, 3])
    if tar == 'rot6d':
      rot6d = np.reshape(
        np.transpose(rot[:, :, :2], [0, 2, 1]), data_shape + (6,)
      )
      return rot6d

    if tar == 'axangle':
      angle = np.arccos((rot[:, 0, 0] + rot[:, 1, 1] + rot[:, 2, 2] - 1) / 2)
      angle = np.expand_dims(angle, -1)
      norm = np.sqrt(
        np.square(rot[:, 2, 1] - rot[:, 1, 2]) + \
        np.square(rot[:, 0, 2] - rot[:, 2, 0]) + \
        np.square(rot[:, 1, 0] - rot[:, 0, 1])
      )
      norm = np.maximum(norm, np.finfo(np.float32).eps)
      x = (rot[:, 2, 1] - rot[:, 1, 2]) / norm
      y = (rot[:, 0, 2] - rot[:, 2, 0]) / norm
      z = (rot[:, 1, 0] - rot[:, 0, 1]) / norm
      axangle = np.stack([x, y, z], -1) * angle
      axangle = np.reshape(axangle, data_shape + (3,))
      return axangle

    if tar == 'quat':
      quat = []
      for i in range(rot.shape[0]):
        tr = rot[i, 0, 0] + rot[i, 1, 1] + rot[i, 2, 2]
        if tr > 0:
          S = np.sqrt(tr + 1.0) * 2
          qw = 0.25 * S
          qx = (rot[i, 2, 1] - rot[i, 1, 2]) / S
          qy = (rot[i, 0, 2] - rot[i, 2, 0]) / S
          qz = (rot[i, 1, 0] - rot[i, 0, 1]) / S
        elif rot[i, 0, 0] > rot[i, 1, 1] and rot[i, 0, 0] > rot[i, 2, 2]:
          S = np.sqrt(1.0 + rot[i, 0, 0] - rot[i, 1, 1] - rot[i, 2, 2]) * 2
          qw = (rot[i, 2, 1] - rot[i, 1, 2]) / S
          qx = 0.25 * S
          qy = (rot[i, 0, 1] + rot[i, 1, 0]) / S
          qz = (rot[i, 0, 2] + rot[i, 2, 0]) / S
        elif rot[i, 1, 1] > rot[i, 2, 2]:
          S = np.sqrt(1.0 + rot[i, 1, 1] - rot[i, 0, 0] - rot[i, 2, 2]) * 2
          qw = (rot[i, 0, 2] - rot[i, 2, 0]) / S
          qx = (rot[i, 0, 1] + rot[i, 1, 0]) / S
          qy = 0.25 * S
          qz = (rot[i, 1, 2] + rot[i, 2, 1]) / S
        else:
          S = np.sqrt(1.0 + rot[i, 2, 2] - rot[i, 0, 0] - rot[i, 1, 1]) * 2
          qw = (rot[i, 1, 0] - rot[i, 0, 1]) / S
          qx = (rot[i, 0, 2] + rot[i, 2, 0]) / S
          qy = (rot[i, 1, 2] + rot[i, 2, 1]) / S
          qz = 0.25 * S
        quat.append(np.array([qw, qx, qy, qz]))
      quat = np.array(quat)
      quat = np.reshape(quat, data_shape + (4,))
      return quat

  if src == 'rot6d':
    data_shape = rot.shape[:-1]
    rot = np.reshape(rot, [-1, 6])
    col0 = rot[:, 0:3] / \
        np.maximum(np.linalg.norm(rot[:, 0:3], axis=-1, keepdims=True), eps)
    col1 = rot[:, 3:6] - np.sum((col0 * rot[:, 3:6]), axis=-1, keepdims=True) * col0
    col1 = col1 / np.maximum(np.linalg.norm(col1, axis=-1, keepdims=True), eps)
    col2 = np.cross(col0, col1)
    rotmat = np.stack([col0, col1, col2], -1)
    rotmat = np.reshape(rotmat, data_shape + (3, 3))
    if tar == 'rotmat':
      return rotmat
    return convert(rotmat, 'rotmat', tar)

  raise NotImplementedError(f'Unsupported conversion: from {src} to {tar}.')


def rotmat_rel_to_abs(rel_rotmat, parents, batch=False):
  if not batch:
    rel_rotmat = np.expand_dims(rel_rotmat, 0)
  n_joints = len(parents)
  abs_rotmat = [None] * n_joints
  for c in range(n_joints):
    abs_rotmat[c] = rel_rotmat[:, c]
    p = parents[c]
    while p is not None:
      abs_rotmat[c] = \
        np.einsum('nhw, nwk -> nhk', rel_rotmat[:, p], abs_rotmat[c])
      p = parents[p]
  abs_rotmat = np.stack(abs_rotmat, 1)
  if not batch:
    abs_rotmat = abs_rotmat[0]
  return abs_rotmat


def rotmat_abs_to_rel(abs_rotmat, parents, batch=False):
  if not batch:
    abs_rotmat = np.expand_dims(abs_rotmat, 0)
  n_joints = len(parents)
  rel_rotmat = [None] * n_joints
  for c in range(n_joints):
    p = parents[c]
    if p is None:
      rel_rotmat[c] = abs_rotmat[:, c]
    else:
      rel_rotmat[c] = np.einsum(
        'nhw, nwd -> nhd',
        np.transpose(abs_rotmat[:, p], (0, 2, 1)), abs_rotmat[:, c]
      )
  rel_rotmat = np.stack(rel_rotmat, 1)
  if not batch:
    rel_rotmat = rel_rotmat[0]
  return rel_rotmat


def keypoints_to_bones(keypoints, parents, batch=False):
  if not batch:
    keypoints = np.expand_dims(keypoints, 0)
  bones = []
  for c, p in enumerate(parents):
    if p is None:
      bones.append(keypoints[:, c])
    else:
      bones.append(keypoints[:, c] - keypoints[:, p])
  bones = np.stack(bones, 1)
  if not batch:
    bones = bones[0]
  return bones


def bones_to_keypoints(bones, parents, batch=False):
  if not batch:
    bones = np.expand_dims(bones)
  keypoints = []
  for c, p in enumerate(parents):
    if p is None:
      keypoints.append(bones[:, c])
    else:
      keypoints.append(bones[:, c] + keypoints[p])
  keypoints = np.stack(keypoints, 1)
  if not batch:
    keypoints = keypoints[0]
  return keypoints


def forward_kinematics(ref_bones, abs_rotmat, parents, batch=False):
  if not batch:
    abs_rotmat = np.expand_dims(abs_rotmat, 0)
    ref_bones = np.expand_dims(ref_bones, 0)
  bones = np.einsum('njhw, njw -> njh', abs_rotmat, ref_bones)
  keypoints = bones_to_keypoints(bones, parents, batch=True)
  if not batch:
    bones = bones[0]
    keypoints = keypoints[0]
  return keypoints, bones


def sphere_sampling(n):
  theta = np.random.uniform(0, 2 * np.pi, size=n)
  phi = np.random.uniform(0, np.pi, size=n)
  v = np.stack(
    [np.cos(theta) * np.sin(phi), np.sin(theta) * np.sin(phi), np.cos(phi)], 1
  )
  return v


def random_rotation(n):
  axis = sphere_sampling(n)
  angle = np.random.uniform(np.pi, size=[n, 1])
  return convert(axis * angle, 'axangle', 'rotmat')


def slerp(a, b, t):
  shape = list(a.shape[:-1])
  a = np.reshape(a, [-1, 4])
  b = np.reshape(b, [-1, 4])
  t = np.reshape(t, [-1, 1])

  dot = np.einsum('nd, nd -> n', a, b)
  omega = np.expand_dims(np.arccos(np.clip(dot, -1, 1), dtype=np.float32), -1)
  so = np.sin(omega, dtype=np.float32)
  so[so == 0] = np.finfo(np.float32).eps
  p = np.sin((1 - t) * omega, dtype=np.float32) / so * a + \
      np.sin(t * omega, dtype=np.float32) / so * b
  mask = np.expand_dims(np.abs(dot) >= 1, -1)
  p = np.where(mask, a, p)

  p = np.reshape(p, shape + [4])
  return p



def procrustes_align(s1, s2, return_rst=False):
  """
  Procrustes alignment. s1_hat = (s * r.dot(s1.T) + t).T

  Parameters
  ----------
  s1 : [N, 3]
    Array to align from.
  s2 : [N, 3]
    Array to align to.

  Returns
  -------
  [N, 3]
    s1 after alignment.
  """
  if s1.shape[0] <= 2:
    return s2.copy();

  s1 = s1.T
  s2 = s2.T

  m1 = s1.mean(axis=1, keepdims=True)
  m2 = s2.mean(axis=1, keepdims=True)
  x1 = s1 - m1
  x2 = s2 - m2

  var1 = np.sum(x1**2)

  K = x1.dot(x2.T)

  U, _, Vh = np.linalg.svd(K)
  V = Vh.T
  Z = np.eye(U.shape[0])
  Z[-1, -1] *= np.sign(np.linalg.det(U.dot(V.T)))
  R = V.dot(Z.dot(U.T))

  scale = np.trace(R.dot(K)) / var1

  t = m2 - scale * (R.dot(m1))

  s1_hat = scale * R.dot(s1) + t

  if return_rst:
    return R, scale, t

  return s1_hat.T
